/* #includes */ /*{{{C}}}*//*{{{*/
#ifndef NO_POSIX_SOURCE
#undef _POSIX_SOURCE
#define _POSIX_SOURCE   1
#undef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 2
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "latex.h"
#include "main.h"
#include "misc.h"
/*}}}*/

/* savelatex     -- save as LaTeX table */ /*{{{*/
const char *savelatex(Sheet *sheet, const char *name, int body,
		      const Location beg, const Location end,
		      unsigned int *count)
{
  assert(sheet != (Sheet*)0);
  assert(name != NULL);
  
  *count=0;
  Location w;
  w[X] = beg[X];
  for (w[Z]=beg[Z]; w[Z]<=end[Z]; ++(w[Z]))
    for (w[Y]=beg[Y]; w[Y]<=end[Y]; ++(w[Y]))
      if (shadowed(sheet, w)) return _("Shadowed cells in first column");
  FILE *fp = (FILE*)0; /* cause runtime error */
  if (!body && (fp=fopen(name,"w"))==(FILE*)0) return strerror(errno);
  for (w[Z]=beg[Z]; w[Z]<=end[Z]; ++(w[Z]))
  {
    if (body)
    /* open new file */ /*{{{*/
    {
      char num[20];
      sprintf(num, ".%d", w[Z]);
      char fullname[PATH_MAX];
      fullname[sizeof(fullname)-strlen(num)-1]='\0';
      (void)strncpy(fullname,name,sizeof(fullname)-strlen(num)-1);
      fullname[sizeof(fullname)-1]='\0';
      (void)strncat(fullname,num,sizeof(fullname)-strlen(num)-1);
      fullname[sizeof(fullname)-1]='\0';  
      if ((fp=fopen(fullname,"w"))==(FILE*)0) return strerror(errno);
    }
    /*}}}*/
    else
    /* print header */ /*{{{*/
    if (w[Z] == beg[Z])
    {
      if (fputs_close("\\documentclass{article}\n\\usepackage{longtable}\n\\begin{document}\n\\setlongtables\n\\vfill\n",fp)==EOF) return strerror(errno);
    }
    else
    {
      if (fputs_close("\\clearpage\n\\vfill\n",fp)==EOF) return strerror(errno);
    }
    /*}}}*/

    /* print bogus format */ /*{{{*/
    fprintf(fp,"\\begin{longtable}{");
    for (w[X]=beg[X]; w[X]<=end[X]; ++(w[X])) if (fputc_close('l',fp)==EOF) return strerror(errno);
    fprintf(fp,"}\n");
    /*}}}*/
    for (w[Y]=beg[Y]; w[Y]<=end[Y]; ++(w[Y]))
    /* print contents */ /*{{{*/
    {
      for (w[X]=beg[X]; w[X]<=end[X]; )
      {
	if (w[X] > beg[X] && fputc_close('&', fp) == EOF)
	  return strerror(errno);
	Location mw; LOCATION_GETS(mw,w);
        for (mw[X]++; (size_t)(mw[X]) < sheet->dim[X] && shadowed(sheet, mw);
	     mw[X]++);
        int multicols = mw[X] - w[X];
        fprintf(fp, "\\multicolumn{%d}{", multicols);
	Style cs = getstyle(sheet, w);
        switch (cs.adjust)
        {
          case LEFT: if (fputc_close('l',fp)==EOF) return strerror(errno); break;
          case RIGHT: if (fputc_close('r',fp)==EOF) return strerror(errno); break;
          case CENTER: if (fputc_close('c',fp)==EOF) return strerror(errno); break;
          default: assert(0);
        }
	char buf[1024];
        printvalue(buf, sizeof(buf), 0, DIRECT_STRING, cs.fform, cs.precision,
		   sheet, w);
        if (fputs_close("}{", fp) == EOF) return strerror(errno);
        if (cs.transparent)
        {
          if (fputs_close(buf,fp)==EOF) return strerror(errno);
        }
        else for (char *bufp = buf; *bufp; ++bufp) switch (*bufp)
        {
          case '%':
          case '$':
          case '&':
          case '#':
          case '_':
          case '{':
          case '}':
          case '~':
          case '^': if (fputc_close('\\',fp)==EOF || fputc_close(*bufp,fp)==EOF) return strerror(errno); break;
          case '\\': if (fputs_close("\\backslash ",fp)==EOF) return strerror(errno); break;
          default: if (fputc_close(*bufp,fp)==EOF) return strerror(errno);
        }
        if (fputc_close('}',fp)==EOF) return strerror(errno);
        w[X] += multicols;
        ++*count;
      }
      if (fputs_close(w[Y]<end[Y] ? "\\\\\n" : "\n\\end{longtable}\n",fp)==EOF) return strerror(errno);
    }
    /*}}}*/
    if (body)
    {
      if (fclose(fp)==EOF) return strerror(errno);
    }
    else
    {
      if (fputs_close("\\vfill\n",fp)==EOF) return strerror(errno);
    }
  }
  if (!body)
  {
    if (fputs_close("\\end{document}\n",fp)==EOF) return strerror(errno);
    if (fclose(fp)==EOF) return strerror(errno);
  }
  return (const char*)0;
}
/*}}}*/
